from .connection import RedisInstance


def is_redis_connected(redis: RedisInstance) -> tuple[bool, Exception | None]:
    try:
        return redis.get_redis_client().ping(), None
    except Exception as e:
        return False, e



from typing import Any
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse, Response


def return_response(
        data: Any,
        status_code: int,
        response_class: Response,
        return_json_directly: bool = False,
        headers: dict[int, dict] = None,
        no_cache: bool = True,
) -> Response:
    if headers is None:
        headers = {}

    final_headers = headers.get(status_code) or {}

    if no_cache:
        final_headers.update({
            'Cache-Control': 'private, no-cache, no-store, must-revalidate, max-age=0, s-maxage=0',
            'Pragma': 'no-cache',
            'Expires': '0',
        })

    if response_class != JSONResponse:
        return_json_directly = True

    if return_json_directly:
        content = data
    else:
        if status_code < 300:
            content = {'success': True, 'data': jsonable_encoder(data)}
        else:
            content = {'success': False, 'message': data}

    return response_class(
        status_code=status_code,
        content=content,
        headers=final_headers
    )


def return_json_response(
        data: Any,
        status_code: int,
        return_json_directly: bool = False,
        headers: dict[int, dict] = None,
        no_cache: bool = True,
) -> JSONResponse:
    return return_response(
        data=data,
        status_code=status_code,
        return_json_directly=return_json_directly,
        headers=headers,
        no_cache=no_cache,
        response_class=JSONResponse,
    )



import boto3
from botocore.client import BaseClient
from botocore.exceptions import ClientError, NoCredentialsError

from ..core.exception import RequestException


class Bucket:
    def __init__(
            self,
            name: str | None = None,
            s3_mode: str | None = None,
            endpoint_url: str | None = None,
            access_key: str | None = None,
            secret_key: str | None = None,
            region_name: str | None = None,
    ):
        self.__name = name
        self.__s3_mode = bool(s3_mode)
        self.__endpoint_url = endpoint_url
        self.__access_key = access_key
        self.__secret_key = secret_key
        self.__region_name = region_name
        self.__s3_client: BaseClient = boto3.client(
            service_name='s3',
            aws_access_key_id=self.__access_key,
            aws_secret_access_key=self.__secret_key,
            endpoint_url=self.__endpoint_url if not self.__s3_mode else None,
            region_name=self.__region_name if self.__s3_mode else None,
        )

    def url_generator(self, key: str):
        if self.__s3_mode:
            return f"{self.__endpoint_url}/{key}"
        else:
            return f"{self.__endpoint_url}/{self.__name}/{key}"

    def get_s3(self) -> BaseClient:
        return self.__s3_client

    def get_bucket_name(self) -> str | None:
        return self.__name

    def check_file_exists(self, key: str):
        try:
            self.__s3_client.head_object(Bucket=self.__name, Key=key)
            return True
        except NoCredentialsError:
            raise RequestException(
                controller=f'bucket.check_file_exists',
                message='S3 credentials error',
                status_code=500,
            )
        except:
            return False

    def upload(self, file: bytes, key: str, content_type: str, cache_control='no-cache'):
        try:
            self.__s3_client.put_object(
                Body=file,
                Bucket=self.__name,
                Key=key,
                ContentType=content_type,
                CacheControl=cache_control,
            )
        except NoCredentialsError:
            raise RequestException(
                controller=f'bucket.check_file_exists',
                message='S3 credentials error',
                status_code=500,
            )
        return self.url_generator(key=key)

    def download(self, key: str, filepath: str):
        try:
            self.__s3_client.download_file(
                Bucket=self.__name,
                Key=key,
                Filename=filepath,
            )
        except NoCredentialsError:
            raise RequestException(
                controller=f'bucket.check_file_exists',
                message='S3 credentials error',
                status_code=500,
            )
        return True

    def download_fileobj(self, key: str, file):
        try:
            self.__s3_client.download_fileobj(
                Bucket=self.__name,
                Key=key,
                Fileobj=file,
            )
        except NoCredentialsError:
            raise RequestException(
                controller=f'bucket.check_file_exists',
                message='S3 credentials error',
                status_code=500,
            )
        return True

    def upload_by_path(self, file_path: str, key: str, content_type: str | None = None, cache_control='no-cache'):
        try:
            with open(file_path, 'rb') as file:
                file_content = file.read()
                content_type = content_type or 'application/octet-stream'
                return self.upload(file=file_content, key=key, content_type=content_type, cache_control=cache_control)
        except NoCredentialsError:
            raise RequestException(
                controller=f'bucket.check_file_exists',
                message='S3 credentials error',
                status_code=500,
            )

    def duplicate(self, source_key: str, destination_key: str, cache_control='no-cache'):
        try:
            self.__s3_client.copy(
                CopySource={
                    'Bucket': self.__name,
                    'Key': source_key
                },
                Bucket=self.__name,
                Key=destination_key,
                ExtraArgs={
                    'CacheControl': cache_control
                }
            )
        except NoCredentialsError:
            raise RequestException(
                controller=f'bucket.check_file_exists',
                message='S3 credentials error',
                status_code=500,
            )
        return self.url_generator(key=destination_key)

    def safe_duplicate(self, source_key: str, cache_control='no-cache'):
        try:
            default_key = source_key
            key = default_key
            i = 2
            while self.check_file_exists(key):
                name, ext = default_key.rsplit(".", 1)
                key = f"{name}-{i}.{ext}"
                i += 1
        except NoCredentialsError:
            raise RequestException(
                controller=f'bucket.check_file_exists',
                message='S3 credentials error',
                status_code=500,
            )

        return self.duplicate(
            source_key=source_key,
            destination_key=key,
            cache_control=cache_control
        )

    def delete(self, key: str):
        try:
            self.__s3_client.delete_object(Bucket=self.__name, Key=key)
            return True
        except NoCredentialsError:
            raise RequestException(
                controller=f'bucket.check_file_exists',
                message='S3 credentials error',
                status_code=500,
            )
        except ClientError:
            return True



import os


def make_directory(path):
    if not os.path.exists(path):
        os.makedirs(path)


def folder_path_of_file(path):
    return os.path.dirname(os.path.realpath(path))


def remove_file(path):
    if os.path.exists(path):
        os.remove(path)



import logging


def leave(
        log_type: str,
        message: str | None = None,
        **kwargs,
):
    logger = logging.getLogger()

    kwargs['footprint'] = True
    kwargs['message'] = message

    data = dict(
        msg=message,
        extra={
            'details': kwargs,
        }
    )

    error_mapper = {
        'error': logger.error,
        'warning': logger.warning,
        'debug': logger.debug
    }.get(log_type.lower(), logger.info)

    error_mapper(**data)



def safe_access(func, default_value=None):
    try:
        return func()
    except:
        return default_value


def convert_to_int_or_float(string_num):
    try:
        float_num = float(string_num)
        if float_num.is_integer():
            return int(float_num)
        else:
            return float_num
    except ValueError:
        return None



import traceback
import sys


class RequestException(Exception):
    def __init__(self, status_code=500, controller=None, message='', skip_footprint: bool = False):
        self.status_code = status_code
        self.controller = controller
        self.message = message
        self.skip_footprint = skip_footprint
        super().__init__(self.controller)


def exception_to_dict(exc):
    exc_type, exc_obj, tb = sys.exc_info()

    exc_dict = {
        'type': str(exc_type.__name__),
        'message': str(exc),
    }

    tb_info = traceback.extract_tb(tb)
    detailed_tb = []
    for tb in tb_info:
        tb_details = {
            'filename': tb.filename,
            'line': tb.lineno,
            'function': tb.name,
            'text': tb.line,
        }
        detailed_tb.append(tb_details)

    exc_dict['traceback'] = detailed_tb

    if hasattr(exc, 'args'):
        exc_dict['args'] = exc.args

    return exc_dict



def chunk_list(lst, chunk_size):
    """Splits a list into chunks of a specified size."""
    return [lst[i:i + chunk_size] for i in range(0, len(lst), chunk_size)]


def chunk_list_generator(lst, chunk_size):
    """Generates chunks of a specified size from a list."""
    for i in range(0, len(lst), chunk_size):
        yield lst[i:i + chunk_size]



from urllib.parse import urlencode, urlparse, parse_qs, urlunparse


def add_query_param(url, params):
    url_parts = urlparse(url)
    query_params = parse_qs(url_parts.query)
    query_params.update(params)
    new_query = urlencode(query_params, doseq=True)
    new_url = urlunparse((
        url_parts.scheme,
        url_parts.netloc,
        url_parts.path,
        url_parts.params,
        new_query,
        url_parts.fragment
    ))
    return new_url



import json
from dataclasses import dataclass

from ..core.retry import retry_wrapper

from .connection import RedisInstance


@dataclass
class Message:
    name: str
    body: dict

    def get_json_encoded(self):
        return {
            'name': self.name,
            'body': json.dumps(self.body, default=str)
        }


class Sender:

    def __init__(self, redis_instance: RedisInstance):
        self._channels = []
        self._redis_instance = redis_instance
        self._redis_client = self._redis_instance.get_redis_client()

    def register_channel(self, channel_name: str):
        self._channels.append(channel_name)

    @retry_wrapper()
    def send_message(self, channel: str, message: Message):
        self._redis_client.xadd(channel, message.get_json_encoded())



import hashlib
import json
import zlib
from typing import Any
from functools import wraps
from redis import Redis

from ..core.exception import exception_to_dict
from ..core.jsonable_encoder import jsonable_encoder
from ..log import footprint

from .connection import RedisInstance


def get_data_directly(
        func,
        *args,
        **kwargs
):
    try:
        return func(*args, **kwargs)
    except Exception as e:
        footprint.leave(
            log_type='error',
            message='Error occurred while executing the function.',
            controller='caching.get_data_directly',
            subject='Error on running a function',
            payload=exception_to_dict(e),
        )
        raise e


def cache_data(
        response: dict,
        cache_key: str,
        redis_instance: Redis,
        expire: int | None = None
):
    try:
        compressed_main_value = zlib.compress(json.dumps(jsonable_encoder(response)).encode('utf-8'))
        redis_instance.delete(cache_key)
        if expire:
            redis_instance.setex(name=cache_key, value=compressed_main_value, time=expire)
        else:
            redis_instance.set(name=cache_key, value=compressed_main_value)

    except Exception as exception:
        footprint.leave(
            log_type='error',
            message='We faced an error while we want to cache data.',
            controller='caching.cache_data',
            subject='Error on caching data',
            payload={
                'expire': expire,
                'cache_key': cache_key,
                'error': exception_to_dict(exception),
            },
        )

    return response


def cache_function(
    func: callable,
    redis: RedisInstance,
    namespace: str,
    expire: int | None = None,
    cache_only_for: list[dict[str, Any]] = None,
    skip_cache_keys: set[str] = None,
    *args,
    **kwargs,
):
    """
    Call a function with caching support. If a cached value exists in Redis, it is returned.
    Otherwise, the function is executed, its result is cached, and then returned.

    :param redis: An instance of redis class.
    :param skip_cache_keys: The keys that we should not involve them in caching:
    :param func: The function to be called.
    :param namespace: A string used as a namespace in the cache key.
    :param expire: Optional expiration time (in seconds) for the cache entry.
    :param cache_only_for: The situation that caching should be done.
    :param args: Positional arguments for `func`.
    :param kwargs: Keyword arguments for `func`.
    :return: The result of the function call, possibly validated via the response_model.
    """
    if skip_cache_keys is None:
        skip_cache_keys = set()

    if cache_only_for is not None:
        should_be_cached = False
        for cache_condition in cache_only_for:
            cache_col = cache_condition.get('kwarg')
            if cache_condition.get('operator') == 'in' and kwargs.get(cache_col) in cache_condition.get('value'):
                should_be_cached = True
                break

        if not should_be_cached:
            return get_data_directly(func=func, *args, **kwargs)

    # Extract a website value if provided and remove potential non-deterministic keys.
    kwargs_key = {k: v for k, v in kwargs.items() if k not in skip_cache_keys}

    # Build a deterministic cache key.
    cache_key = ""
    if namespace:
        cache_key += f"{namespace}:"
    if args:
        args_hash = hashlib.sha256(json.dumps(args, default=str).encode('utf-8')).hexdigest()
        cache_key += f"{args_hash}:"
    if kwargs_key:
        kwargs_hash = hashlib.sha256(json.dumps(kwargs_key, default=str).encode('utf-8')).hexdigest()
        cache_key += f"{kwargs_hash}:"
    cache_key = cache_key.rstrip(':')

    with redis.get_redis_client() as redis_instance:
        # Attempt to retrieve cached data.
        try:
            cache_compressed = redis_instance.get(cache_key)
        except Exception as exception:
            footprint.leave(
                log_type='error',
                message='Error while trying to retrieve data from cache.',
                controller='caching.cache_function',
                subject='Error on get cached data',
                payload={
                    'redis_key': cache_key,
                    'error': exception_to_dict(exception)
                },
            )
            cache_compressed = None

        if cache_compressed is not None:
            try:
                response = json.loads(zlib.decompress(cache_compressed).decode('utf-8'))
                return response
            except Exception as exception:
                footprint.leave(
                    log_type='error',
                    message='Error during decompressing or loading cached data.',
                    controller='caching.cache_function',
                    subject='Error on reading cache',
                    payload={
                        'redis_key': cache_key,
                        'error': exception_to_dict(exception)
                    },
                )

        # If cache miss or error in retrieving cache, execute the function.
        result = get_data_directly(func=func, *args, **kwargs)

        # Cache the result.
        try:
            compressed_value = zlib.compress(json.dumps(jsonable_encoder(result)).encode('utf-8'))
            redis_instance.delete(cache_key)
            if expire:
                redis_instance.setex(name=cache_key, value=compressed_value, time=expire)
            else:
                redis_instance.set(name=cache_key, value=compressed_value)
        except Exception as exception:
            footprint.leave(
                log_type='error',
                message='Error occurred while caching the result.',
                controller='caching.cache_function',
                subject='Error on reading cache',
                payload={
                    'redis_key': cache_key,
                    'error': exception_to_dict(exception)
                },
            )
        return result


def cache_wrapper(
    redis: RedisInstance,
    namespace: str,
    expire: int | None = None,
    cache_only_for: list[dict[str, Any]] = None,
    skip_cache_keys: set[str] = None,
):
    def decorator(func: callable):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return cache_function(
                func=func,
                redis=redis,
                namespace=namespace,
                expire=expire,
                cache_only_for=cache_only_for,
                skip_cache_keys=skip_cache_keys,
                *args,
                **kwargs,
            )
        return wrapper

    return decorator



import logging
from logging.handlers import RotatingFileHandler

from .api_handler import LoggerHandler
from .config import LogConfig
from .formatter import CustomFormatter


def get_handlers_data(config: LogConfig):
    formatter = CustomFormatter()

    logging_api_url = config.get('api_url')
    logging_api_key = config.get('api_key')
    only_footprint_mode = config.get('only_footprint_mode', True)

    log_print = config.get('log_print', default=False)
    log_store = config.get('log_store', default=False)
    log_level = getattr(logging, config.get('log_level', default='INFO'))

    handlers = []

    if logging_api_url and logging_api_key:
        api_handler = LoggerHandler(
            logging_api_url=logging_api_url,
            logging_api_key=logging_api_key,
            only_footprint_mode=only_footprint_mode,
        )
        api_handler.setLevel(log_level)
        api_handler.setFormatter(formatter)
        handlers.append(api_handler)

    if log_print:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(log_level)
        console_handler.setFormatter(formatter)
        handlers.append(console_handler)

    if log_store:
        log_file_name = config.get('log_file_name')
        log_file_backup_count = config.get('log_file_backup_count')
        log_file_max_size = config.get('log_file_max_size')
        rotating_handler = RotatingFileHandler(log_file_name, maxBytes=log_file_max_size, backupCount=log_file_backup_count)
        rotating_handler.setLevel(log_level)
        rotating_handler.setFormatter(formatter)
        handlers.append(rotating_handler)

    return handlers, log_level



from sqlalchemy import text

from .database import DatabaseInstance


def is_database_connected(db: DatabaseInstance) -> tuple[bool, Exception | None]:
    try:
        with db.engine_write.connect() as connection:
            connection.execute(text("SELECT 1"))

        with db.engine_read.connect() as connection:
            connection.execute(text("SELECT 1"))

        return True, None

    except Exception as e:
        return False, e



from typing import Any
from math import ceil
from datetime import datetime
from sqlalchemy import func, or_, cast, and_, union, String, distinct as distinct_func
from sqlalchemy.orm import Session


def convert_symbols_to_unicode(string: str):
    cleaned_string = ''
    for character in string:
        if ord(character) > 127:
            cleaned_string += '%'
        else:
            cleaned_string += character
    return cleaned_string


def filters_mapper(db: Session, pre_conditions: list, joins: list[dict], filter_item, conditions):
    columns = filter_item.get('columns')
    columns_type = filter_item.get('type', 'select')

    if columns_type == 'select':
        distinct_names_query = []
        for column in columns:
            query = db.query(
                func.jsonb_array_elements_text(column)
                if filter_item.get('is_json', False) else
                column
            )

            for join_item in joins:
                query = query.join(**join_item)

            distinct_names_query.append(query.filter(*pre_conditions, *conditions).distinct())

        values = list(map(lambda x: x[0], db.execute(union(*distinct_names_query)).fetchall()))
        enum = filter_item.get('enum')
        labels = filter_item.get('labels', {})
        items = [
            {
                'label': labels.get(enum(value) if enum else value, value),
                'value': enum(value).name if enum else value,
            }
            for value in values
            if value is not None
        ]
        items = sorted(items, key=lambda x: x.get('value'))
    elif columns_type in ['number', 'date']:
        main_query = db.query(
            func.least(*list(map(func.min, columns))),
            func.greatest(*list(map(func.max, columns))),
        )
        for join_item in joins:
            main_query = main_query.join(**join_item)

        min_value, max_value = main_query.filter(*pre_conditions, *conditions).first()
        items = {
            'min': min_value,
            'max': max_value
        }
    else:
        items = None

    return items


def get_filters_value(db: Session, pre_conditions: list, joins: list[dict], filters: list, names_conditions: dict):
    available_filters = []
    for filter_item in filters:
        if not filter_item.get('hide_in_response', False):
            filter_name_conditions = names_conditions.get(filter_item['name'])
            if filter_name_conditions is None:
                continue
            filter_mapping_result = filters_mapper(
                db=db,
                pre_conditions=pre_conditions,
                joins=joins,
                filter_item=filter_item,
                conditions=filter_name_conditions
            )
            filter_label = filter_item.get('label')
            filter_type = filter_item.get('type')
            filter_name = filter_item.get('name')
            if filter_mapping_result is not None:
                if filter_type == 'select':
                    available_filters.append({
                        'label': filter_label,
                        'name': filter_name,
                        'type': filter_type,
                        'value': filter_mapping_result
                    })
                elif filter_type in ['number', 'date']:
                    available_filters.append({
                        'label': filter_label,
                        'name': filter_name,
                        'type': filter_type,
                        'value': filter_mapping_result,
                    })
                else:
                    available_filters.append({
                        'label': filter_label,
                        'name': filter_name,
                        'type': filter_type,
                    })

    return available_filters


def make_selected_filters(
        filters: list = None,
        current_query: dict[str, Any] = None,
):
    selected_filters = []
    for filter_item in filters:
        selected_data: dict | list | None = current_query.get(filter_item['name'])
        if not selected_data or len(selected_data) == 0:
            continue

        if isinstance(selected_data, dict) and len([k for k, v in selected_data.items() if v is not None]) == 0:
            continue

        if filter_item['type'] == 'select':
            for data in selected_data:
                enum = filter_item.get('enum')
                labels = filter_item.get('labels', {})
                if enum:
                    data = getattr(enum, str(data), data)

                selected_filters.append({
                    'label': labels.get(enum(data) if enum else data, data),
                    'name': filter_item['name'],
                    'value': enum(data).name if enum else data,
                    'type': filter_item['type']
                })
        elif filter_item['type'] == 'number':
            minimum = selected_data.get('min')
            maximum = selected_data.get('max')
            if minimum is not None and maximum is not None:
                result_label = f"{filter_item['label']} (between {minimum} and {maximum})"
            elif minimum is None and maximum is not None:
                result_label = f"{filter_item['label']} (To {maximum})"
            elif minimum is not None and maximum is None:
                result_label = f"{filter_item['label']} (From {minimum})"
            else:
                result_label = filter_item['label']

            selected_filters.append({
                'label': result_label,
                'name': filter_item['name'],
                'value': selected_data,
                'type': filter_item['type']
            })
        elif filter_item['type'] == 'date':
            minimum = selected_data.get('min')
            maximum = selected_data.get('max')
            if minimum is not None and maximum is not None:
                result_label = f"{filter_item['label']} (From {minimum} To {maximum})"
            elif minimum is None and maximum is not None:
                result_label = f"{filter_item['label']} (To {maximum})"
            elif minimum is not None and maximum is None:
                result_label = f"{filter_item['label']} (From {minimum})"
            else:
                result_label = filter_item['label']

            selected_filters.append({
                'label': result_label,
                'name': filter_item['name'],
                'value': selected_data,
                'type': filter_item['type']
            })
        elif filter_item['type'] == 'search':
            selected_filters.append({
                'label': f"Search ({selected_data})",
                'name': filter_item['name'],
                'value': selected_data,
                'type': filter_item['type']
            })

    return selected_filters


def make_condition(
        filter_item: dict,
        values: dict | list | str
):
    columns = filter_item.get('columns')
    columns_logic = filter_item.get('columns_logic', 'or')
    columns_type = filter_item.get('type', 'select')
    is_json_column = filter_item.get('is_json', False)
    if is_json_column:
        columns_type = 'search'

    logic_function = {
        'or': or_,
        'and': and_,
    }.get(columns_logic)

    if columns_type == 'search':
        if is_json_column:
            filter_value = [convert_symbols_to_unicode(w) for w in values]
            return logic_function(
                *[
                    or_(*[func.lower(cast(column, String)).like(func.lower(f"%\"{x}\"%")) for x in filter_value])
                    for column in columns
                ]
            )
        else:
            filter_value = ''.join([c for c in values if ord(c) <= 127])
            return logic_function(
                *[
                    func.lower(cast(column, String)).like(func.lower(f"%{filter_value}%"))
                    for column in columns
                ]
            )
    elif columns_type == 'select':
        return logic_function(
            *[
                getattr(column, 'in_')([
                    convert_symbols_to_unicode(element)
                    if isinstance(element, str) else element
                    for element in values
                ])
                for column in columns
            ]
        )
    elif columns_type == 'number':
        value_min = values.get('min')
        value_max = values.get('max')
        if value_min is not None and value_max is not None:
            return logic_function(
                *[
                    getattr(column, 'between')(value_min, value_max)
                    for column in columns
                ]
            )
        elif value_min is None and value_max is not None:
            return logic_function(
                *[
                    column <= value_max
                    for column in columns
                ]
            )
        elif value_min is not None and value_max is None:
            return logic_function(
                *[
                    column >= value_min
                    for column in columns
                ]
            )

    elif columns_type == 'date':
        value_min: datetime = values.get('min')
        value_max: datetime = values.get('max')

        if value_min is not None and value_max is not None:
            return logic_function(
                *[
                    getattr(column, 'between')(value_min, value_max)
                    for column in columns
                ]
            )
        elif value_min is None and value_max is not None:
            return logic_function(
                *[
                    or_(
                        column <= value_max,
                        column == None
                    )
                    for column in columns
                ]
            )
        elif value_min is not None and value_max is None:
            return logic_function(
                *[
                    column >= value_min
                    for column in columns
                ]
            )

    else:
        return None


def get_list(
        current_query: dict,
        db: Session,
        model: Any,
        joins: list[dict] = None,
        pre_conditions: list = None,
        filters: list = None,
        options: list = None,
        distinct=None,
        primary_column: str = 'id',
        get_function_parameters: dict = None,
        export_mode: bool = False,
):
    if joins is None:
        joins = []

    if filters is None:
        filters = []

    page = current_query.get('page') or 1
    items_per_page = current_query.get('items_per_page') or 30

    if pre_conditions is None:
        pre_conditions = []

    if options is None:
        options = []

    if get_function_parameters is None:
        get_function_parameters = {}

    # Create Initial Model Query
    main_query = db.query(model)

    if distinct is True:
        count_query = db.query(func.count(distinct_func(getattr(model, primary_column))))
    elif distinct:
        count_query = db.query(func.count(distinct_func(distinct)))
    else:
        count_query = db.query(func.count(getattr(model, primary_column)))

    for join_item in joins:
        main_query = main_query.join(**join_item)
        count_query = count_query.join(**join_item)

    main_query = main_query.filter(*pre_conditions)
    count_query = count_query.filter(*pre_conditions)

    # Initialize rows and conditions
    conditions = []

    names_conditions = {}
    for filter_item in filters:
        names_conditions[filter_item.get('name')] = []

    for filter_item in filters:
        name = filter_item.get('name')
        columns = filter_item.get('columns')
        values = current_query.get(name)
        if not columns or not name or not values:
            continue

        target_condition = make_condition(
            filter_item=filter_item,
            values=values
        )
        if target_condition is not None:
            conditions.append(target_condition)
            for inner_name, inner_name_values in names_conditions.items():
                if inner_name != name:
                    inner_name_values.append(target_condition)

    if conditions:
        main_query = main_query.filter(*conditions)
        count_query = count_query.filter(*conditions)

    if distinct is True:
        main_query = main_query.distinct()
    elif distinct:
        main_query = main_query.distinct(distinct)

    if options:
        main_query = main_query.options(*options)

    if sorting := (current_query.get('sorting') or []):
        order_by_list = []
        for item in sorting:
            sort_by = item.get('sort_by')
            if not sort_by:
                continue

            if hasattr(sort_by, 'value'):
                sort_by = sort_by.value

            order_by = item.get('order_by')
            if not order_by:
                continue

            if hasattr(order_by, 'value'):
                order_by = order_by.value

            order_by_list.append(getattr(getattr(model, sort_by), order_by)())

        final_query = main_query.order_by(*order_by_list)
    else:
        final_query = main_query

    if items_per_page and page:
        final_query = (
            final_query
            .limit(items_per_page)
            .offset((page - 1) * items_per_page)
        )

    rows = list(map(lambda x: x.get(**get_function_parameters), final_query.all()))

    if export_mode:
        return rows
    else:
        count = count_query.first()[0]
        current_query['total_row'] = count

        # Calculate pagination-related information
        if items_per_page and page:
            last_page = ceil(count / items_per_page)
            current_query['last_page'] = last_page
            current_query['has_next'] = last_page > page
            current_query['page'] = page
            current_query['items_per_page'] = items_per_page

        # Return a dictionary containing the filter/sort options, current query data, and rows of data
        return {
            'rows_data': rows,
            'payload': current_query,
            'selected_filters': make_selected_filters(
                current_query=current_query,
                filters=filters,
            ),
            'available_filters': get_filters_value(
                db=db,
                pre_conditions=pre_conditions,
                joins=joins,
                filters=filters,
                names_conditions=names_conditions,
            ) if filters is not None else [],
        }



import logging


class CustomFormatter(logging.Formatter):
    def __init__(self):
        super().__init__(None, '%Y-%m-%d %H:%M:%S')

    def format(self, record):
        if hasattr(record, 'details'):
            self._style._fmt = '%(asctime)s - %(levelname)s - %(details)s'
        else:
            self._style._fmt = '%(asctime)s - %(levelname)s - %(message)s'
        return super().format(record)



from fastapi import FastAPI
from fastapi.exceptions import RequestValidationError
from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.middleware.base import BaseHTTPMiddleware

from ..middlewares import runtime, http_exception, validation_exception
from ..routes.router import Router


class SubApp:
    def __init__(
        self,
        application_title: str,
        version: str = "*",
        redoc_url: str = "/",
        docs_url: str = "/swagger",
        middlewares: list = None,
        routers: list[Router] = None,
        gzip_compression: int = None,
        session_middleware_settings: dict = None,
    ):
        """
        Initialize a SubApp instance and create the FastAPI application.
        """
        self.application_title = application_title
        self.version = version
        self.redoc_url = redoc_url
        self.docs_url = docs_url
        self.middlewares = middlewares or []
        self.routers = routers or []
        self.gzip_compression = gzip_compression
        self.session_middleware_settings = session_middleware_settings or {}
        self.app = self._create_app()

    def _create_app(self) -> FastAPI:
        """
        Create and configure a FastAPI application.
        """
        app = FastAPI(
            title=self.application_title,
            version=self.version,
            redoc_url=self.redoc_url,
            docs_url=self.docs_url,
            swagger_ui_parameters={"defaultModelsExpandDepth": -1},
        )

        # Add GZip middleware
        if self.gzip_compression is not None:
            from fastapi.middleware.gzip import GZipMiddleware
            app.add_middleware(GZipMiddleware, minimum_size=self.gzip_compression)

        # Add custom middlewares
        applicable_middlewares = [runtime.Runtime()] + self.middlewares
        for middleware in applicable_middlewares:
            app.add_middleware(BaseHTTPMiddleware, dispatch=middleware)

        # Add session middleware if settings provided
        if self.session_middleware_settings:
            from starlette.middleware.sessions import SessionMiddleware
            app.add_middleware(
                SessionMiddleware,
                **self.session_middleware_settings
            )

        # Exception handlers
        @app.exception_handler(StarletteHTTPException)
        async def http_exception_handler(request, exc):
            return await http_exception.http_exception_handler(request, exc)

        @app.exception_handler(RequestValidationError)
        async def validation_exception_handler(request, exc):
            return await validation_exception.validation_exception_handler(request, exc)

        # Add routers
        for router in self.routers:
            app.include_router(router=router.get_router())

        return app

    def get_app(self) -> FastAPI:
        """
        Get the created FastAPI application instance.
        """
        return self.app



import requests

from ..log import footprint

from .jsonable_encoder import jsonable_encoder
from .exception import RequestException, exception_to_dict


def request(
        method: str,
        path: str,
        host: str | None = None,
        auth_key: str | None = None,
        auth_value: str | None = None,
        auth_type: str | None = None,
        disable_caching: bool = True,
        full_return: bool = False,
        json_return: bool = True,
        internal_service: bool = True,
        **kwargs,
):
    if host is None:
        return None

    if headers := kwargs.get('headers'):
        kwargs['headers'] = jsonable_encoder(headers)

    error_details = {
        'subject': 'Error on sending request',
        'controller': 'dtpyfw.request.request',
        'footprint': True,
        'payload': {
            'host': host,
            'auth_key': auth_key,
            'auth_type': auth_type,
            'method': method,
            'path': path,
            'disable_caching': disable_caching,
            'json_return': json_return,
            'internal_service': internal_service,
        }
    }

    url = f'{host}/{path}'

    if auth_key is not None and auth_value is not None and auth_type is not None:
        if auth_type == 'params':
            if 'params' not in kwargs:
                kwargs['params'] = {}

            kwargs['params'][auth_key] = auth_value
        elif auth_type == 'headers':
            if 'headers' not in kwargs:
                kwargs['headers'] = {}

            kwargs['headers'][auth_key] = auth_value

    if disable_caching:
        if 'headers' not in kwargs:
            kwargs['headers'] = {}

        kwargs['headers'].update({
            'Cache-Control': 'private, no-cache, no-store, must-revalidate, max-age=0, s-maxage=0',
            'Pragma': 'no-cache',
            'Expires': '0',
        })

    try:
        response = requests.request(method=method, url=url, **kwargs)
        status_code = response.status_code
    except Exception as exc:
        error_details['payload']['error'] = exception_to_dict(exc)
        footprint.leave(
            **error_details,
            log_type='error',
            message="We faced an error while we wanted to send an internal request.",
        )
        raise RequestException(
            status_code=500,
            message='Request Sending Error',
            controller='dtpyfw.request.request',
            skip_footprint=True,
        )

    if full_return:
        return response

    elif json_return:
        try:
            response_json = response.json()
        except Exception as exc:
            error_details['payload']['error'] = exception_to_dict(exc)
            error_details['payload']['headers'] = response.headers
            error_details['payload']['text'] = response.text
            footprint.leave(
                **error_details,
                log_type='error',
                message="We faced a non JSON response from the request.",
            )
            raise RequestException(
                status_code=status_code,
                message='Request Parsing Error',
                controller='dtpyfw.request.request',
                skip_footprint=True,
            )
    else:
        return response.text

    if internal_service:
        is_success = response_json.get('success', False) if isinstance(response_json, dict) else False
        if is_success:
            return response_json.get('data')
        else:
            error_details['payload']['headers'] = response.headers
            error_details['payload']['response'] = response_json
            footprint.leave(
                **error_details,
                log_type='error',
                message="We received an error after getting response from the source.",
            )
            raise RequestException(
                status_code=status_code,
                message=response_json.get('message'),
                controller='dtpyfw.request.request',
                skip_footprint=True,
            )
    else:
        return response_json



from jose import jwt
from datetime import datetime, timedelta

from ..core.jsonable_encoder import jsonable_encoder


def jwt_encrypt(
        tokens_secret_key: str,
        encryption_algorithm: str,
        subject: str,
        claims: dict,
        expiration_timedelta: timedelta = None,
):
    data = {
        'subject': subject,
    }
    if expiration_timedelta:
        data['exp'] = (datetime.now() + expiration_timedelta).timestamp()

    data.update(claims)
    return jwt.encode(
        claims=jsonable_encoder(data),
        key=tokens_secret_key,
        algorithm=encryption_algorithm,
    )


def jwt_decrypt(
        tokens_secret_key: str,
        encryption_algorithm: str,
        token: str,
        subject: str,
        check_exp: bool = True,
):
    options = {}
    if check_exp:
        options['require_exp'] = True
        options['verify_exp'] = True
    else:
        options['require_exp'] = False
        options['verify_exp'] = False

    decoded_token = jwt.decode(
        token=token,
        key=tokens_secret_key,
        algorithms=encryption_algorithm,
        options=options if options else None
    )

    if decoded_token.get('subject') != subject:
        raise Exception('wrong_token_subject')

    return decoded_token
